# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Tests for :mod:`igwn_auth_utils`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Duncan Brown, Leo Singer"
__license__ = "BSD-3-Clause"

from .error import IgwnAuthError
from .scitokens import (
    find_token as find_scitoken,
    token_authorization_header as scitoken_authorization_header,
)
from .x509 import (
    find_credentials as find_x509_credentials,
)

try:  # parse version
    from ._version import version as __version__
except ModuleNotFoundError:  # development mode
    __version__ = ''



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

import inspect
import re
import sys
from pathlib import Path

import igwn_auth_utils

# -- metadata ---------------

project = "igwn-auth-utils"
copyright = "2021 Cardiff University"
author = "Duncan Macleod"
top_module = igwn_auth_utils
release = top_module.__version__
version = release.split('.dev', 1)[0]
git_url = "https://github.com/duncanmmacleod/{}".format(project)

# parse version number to get git reference
_setuptools_scm_version_regex = re.compile(
    r"\+g(\w+)(?:\Z|\.)",
)
if match := _setuptools_scm_version_regex.search(release):
    git_ref, = match.groups()
else:
    git_ref = 'v{}'.format(version)

# -- sphinx config ----------

needs_sphinx = "4.0"
extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx_automodapi.automodapi",
]
default_role = "obj"

# -- theme options ----------

html_theme = "sphinx_material"
html_sidebars = {
    "**": [
        "logo-text.html",
        "globaltoc.html",
        "localtoc.html",
        "searchbox.html",
    ],
}
html_theme_options = {
    "repo_url": git_url,
    "repo_name": project,
    "master_doc": False,
    "nav_links": [],
}
html_static_path = ["_static"]

# -- extensions -------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "requests": ("https://docs.python-requests.org/en/stable/", None),
    "requests-gracedb": (
        "https://requests-gracedb.readthedocs.io/en/stable/",
        None,
    ),
}

automodapi_inherited_members = False


def linkcode_resolve(domain, info):
    """Determine the URL corresponding to Python object

    This code is stolen with thanks from the scipy team.
    """
    if domain != "py" or not info["module"]:
        return None

    def find_source(module, fullname):
        obj = sys.modules[module]
        for part in fullname.split("."):
            obj = getattr(obj, part)
        try:  # unwrap a decorator
            obj = obj.im_func.func_closure[0].cell_contents
        except (AttributeError, TypeError):
            pass
        # get filename
        filename = Path(inspect.getsourcefile(obj)).relative_to(
            Path(top_module.__file__).parent,
        ).as_posix()
        # get line numbers of this object
        source, lineno = inspect.getsourcelines(obj)
        if lineno:
            return "{}#L{:d}-L{:d}".format(
                filename,
                lineno,
                lineno + len(source) - 1,
            )
        return filename

    try:
        fileref = find_source(info["module"], info["fullname"])
    except (
        AttributeError,  # object not found
        OSError,  # file not found
        TypeError,  # source for object not found
        ValueError,  # file not
    ):
        raise
        return None

    return "{}/tree/{}/{}/{}".format(
        git_url,
        git_ref,
        top_module.__name__,
        fileref,
    )


# -- setup ------------------

def setup(app):
    app.add_css_file("css/igwnauthutils.css")



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class IgwnAuthError(RuntimeError):
    """Error in discovering/validating an IGWN auth credential.
    """



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Tests for :mod:`igwn_auth_utils.scitokens`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import os
import time
from unittest import mock

from scitokens import SciToken
from scitokens.utils.errors import NonHTTPSIssuer

import pytest

from .. import scitokens as igwn_scitokens
from ..error import IgwnAuthError

ISSUER = "local"
AUDIENCE = "igwn_auth_utils"
_SCOPE_PATH = "/igwn_auth_utils"
READ_SCOPE = "read:{}".format(_SCOPE_PATH)
WRITE_SCOPE = "write:{}".format(_SCOPE_PATH)


def _create_token(
    key=None,
    iss=ISSUER,
    aud=AUDIENCE,
    scope=READ_SCOPE,
    **kwargs
):
    """Create a token
    """
    if key:
        from scitokens.utils.keycache import KeyCache
        keycache = KeyCache.getinstance()
        keycache.addkeyinfo(iss, "test_key", key.public_key())
    now = int(time.time())
    token = SciToken(key=key, key_id="test_key")
    token.update_claims({
        "iat": now,
        "nbf": now,
        "exp": now + 86400,
        "iss": iss,
        "aud": aud,
        "scope": scope,
    })
    token.update_claims(kwargs)
    return token


def _write_token(token, path):
    with open(path, "wb") as file:
        file.write(token.serialize(lifetime=86400))


@pytest.fixture
def rtoken(private_key):
    return _create_token(
        key=private_key,
        scope=READ_SCOPE,
    )


@pytest.fixture
def wtoken(private_key):
    return _create_token(
        key=private_key,
        scope=WRITE_SCOPE,
    )


@pytest.fixture
def rtoken_path(rtoken, tmp_path):
    token_path = tmp_path / "token.use"
    _write_token(rtoken, token_path)
    return token_path


@pytest.fixture
def condor_creds_path(rtoken, wtoken, tmp_path):
    for token, name in (
        (rtoken, "read.use"),
        (wtoken, "write.use"),
    ):
        _write_token(token, tmp_path / name)
    return tmp_path


# -- test utilities -----------------------------

def assert_tokens_equal(a, b):
    _assert_claims_equal(a, b)
    _assert_claims_equal(b, a)


_SKIP_ASSERT_CLAIMS = {
    "exp",
    "iat",
    "nbf",
}


def _assert_claims_equal(a, b):
    for claim, value in a.claims():
        if claim in _SKIP_ASSERT_CLAIMS:
            continue
        assert b.get(claim) == value


# -- tests --------------------------------------

@pytest.mark.parametrize(("scope", "validity"), (
    (READ_SCOPE, True),
    (WRITE_SCOPE, False),
))
def test_is_valid_token(rtoken, scope, validity):
    assert igwn_scitokens.is_valid_token(rtoken, AUDIENCE, scope) is validity


def test_load_token_file(rtoken_path, rtoken, public_pem):
    assert_tokens_equal(
        igwn_scitokens.load_token_file(
            rtoken_path,
            audience=AUDIENCE,
            insecure=True,
            public_key=public_pem,
        ),
        rtoken,
    )


@mock.patch.dict("os.environ")
@pytest.mark.parametrize("envname", (
    "SCITOKEN",
    "BEARER_TOKEN",
))
def test_find_token_env_scitoken(rtoken, public_pem, envname):
    os.environ[envname] = rtoken.serialize(lifetime=86400).decode("utf-8")
    assert_tokens_equal(
        igwn_scitokens.find_token(
            audience=AUDIENCE,
            scope=READ_SCOPE,
            insecure=True,
            public_key=public_pem,
        ),
        rtoken,
    )


@mock.patch.dict("os.environ")
@pytest.mark.parametrize("envname", (
    "SCITOKEN_FILE",
    "BEARER_TOKEN_FILE",
))
def test_find_token_env_scitoken_file(
    rtoken,
    wtoken,
    rtoken_path,
    public_pem,
    envname,
):
    # set the wrong token as SCITOKEN
    os.environ["SCITOKEN"] = wtoken.serialize(lifetime=86400).decode("utf-8")
    # and the correct token as SCITOKEN_FILE
    os.environ[envname] = str(rtoken_path)
    # and make sure we get the correct token back
    assert_tokens_equal(
        igwn_scitokens.find_token(
            audience=AUDIENCE,
            scope=READ_SCOPE,
            insecure=True,
            public_key=public_pem,
        ),
        rtoken,
    )


@mock.patch.dict("os.environ")
def test_find_token_condor_creds(
    rtoken,
    wtoken,
    public_pem,
    condor_creds_path,
):
    os.environ["_CONDOR_CREDS"] = str(condor_creds_path)
    for token, scope in (
        (rtoken, READ_SCOPE),
        (wtoken, WRITE_SCOPE),
    ):
        assert_tokens_equal(
            igwn_scitokens.find_token(
                audience=AUDIENCE,
                scope=scope,
                insecure=True,
                public_key=public_pem,
            ),
            token,
        )


@mock.patch.dict("os.environ")
def test_find_token_error(rtoken, public_pem):
    # token with the wrong claims
    os.environ["SCITOKEN"] = rtoken.serialize().decode("utf-8")
    # check that we get an error
    with pytest.raises(IgwnAuthError) as exc:
        igwn_scitokens.find_token(
            AUDIENCE,
            WRITE_SCOPE,
            insecure=True,
            public_key=public_pem,
        )
    assert str(exc.value).startswith(
        "could not find a valid SciToken",
    )


@mock.patch.dict("os.environ")
@pytest.mark.parametrize(("skip_errors", "error_type", "message"), (
    (False, NonHTTPSIssuer, "Issuer is not over HTTPS"),
    (True, IgwnAuthError, "could not find a valid SciToken"),
))
def test_find_token_skip_errors(rtoken, skip_errors, error_type, message):
    """Check that the ``skip_errors`` keyword for `find_token()` works
    """
    # configure a valid token (wrong claims) **HOWEVER**
    # don't add the necessary keyword arguments that would
    # enable `deserialize` to actually work
    # (see `test_find_token_error` above for that)
    os.environ["SCITOKEN"] = rtoken.serialize().decode("utf-8")

    # check that we get the normal error
    with pytest.raises(error_type) as exc:
        igwn_scitokens.find_token(
            AUDIENCE,
            READ_SCOPE,
            skip_errors=skip_errors,
        )
    assert str(exc.value).startswith(message)


@mock.patch.dict("os.environ")
def test_find_condor_creds_no_env(tmp_path):
    """Check that `_find_condor_creds_token_paths()` handles missing
    ``_CONDOR_CREDS`` environment variable
    """
    assert not list(igwn_scitokens._find_condor_creds_token_paths())


@mock.patch.dict("os.environ")
def test_find_condor_creds_dir_missing(tmp_path):
    """Check that `_find_condor_creds_token_paths()` handles
    ``_CONDOR_CREDS`` pointing at a directory that doesn't exist
    """
    os.environ["_CONDOR_CREDS"] = str(tmp_path / "_condor_creds")
    assert not list(igwn_scitokens._find_condor_creds_token_paths())


@mock.patch.dict("os.environ")
def test_find_condor_creds_dir_empty(tmp_path):
    """Check that `_find_condor_creds_token_paths()` handles
    ``_CONDOR_CREDS`` pointing at an empty directory
    """
    os.environ["_CONDOR_CREDS"] = str(tmp_path)
    assert not list(igwn_scitokens._find_condor_creds_token_paths())


def test_token_authorization_header(rtoken):
    """Check that `token_authorization_header` works
    """
    expected = "Bearer {}".format(rtoken.serialize().decode("utf-8"))

    # reset _serialized_token attribute
    rtoken._serialized_token = None

    # do it once to check that the call to token.serialize() works
    a = igwn_scitokens.token_authorization_header(rtoken)
    assert a == expected

    # do it again to check that the use of _serialized_token attr works
    assert igwn_scitokens.token_authorization_header(rtoken) == a



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Utility functions for discovering valid scitokens
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import os
from pathlib import Path

from .error import IgwnAuthError
from scitokens import (
    Enforcer,
    SciToken,
)

WINDOWS = os.name == "nt"

# -- utilities --------------


def is_valid_token(token, audience, scope, timeleft=600):
    enforcer = Enforcer(token["iss"], audience=audience)

    def _validate_timeleft(value):
        exp = float(value)
        return exp >= enforcer._now + timeleft

    enforcer.add_validator("exp", _validate_timeleft)
    scheme, path = scope.split(":", 1)
    return enforcer.test(token, scheme, path)


# -- I/O --------------------

def deserialize_token(raw, **kwargs):
    """Deserialize a token.

    Parameters
    ----------
    raw : `str`
        the raw serialised token content to deserialise

    **kwargs
        all keyword arguments are passed on to
        :meth:`scitokens.SciToken.deserialize`

    Returns
    -------
    token : `scitokens.SciToken`
        the deserialised token

    See also
    --------
    scitokens.SciToken.deserialize
        for details of the deserialisation, and any valid keyword arguments
    """
    return SciToken.deserialize(raw.strip(), **kwargs)


def load_token_file(path, **kwargs):
    """Load a SciToken from a file path.

    Parameters
    ----------
    path : `str`
        the path to the scitokens file

    **kwargs
        all keyword arguments are passed on to :func:`deserialize_token`

    Returns
    -------
    token : `scitokens.SciToken`
        the deserialised token

    Examples
    --------
    To load a token and validate a specific audience:

    >>> load_token('mytoken', audience="my.service.org")

    See also
    --------
    scitokens.SciToken.deserialize
        for details of the deserialisation, and any valid keyword arguments
    """
    with open(path, "r") as fobj:
        return deserialize_token(fobj.read(), **kwargs)


# -- discovery --------------

def find_token(audience, scope, timeleft=600, skip_errors=False, **kwargs):
    """Find and load a `SciToken` for the given ``audience`` and ``scope``.

    Parameters
    ----------
    audience : `str`
        the required audience (``aud``).

    scope : `str`
        the required scope (``scope``).

    timeleft : `int`
        minimum required time left until expiry (in seconds)
        for a token to be considered 'valid'

    skip_errors : `bool`, optional
        skip over errors encoutered when attempting to deserialise
        discovered tokens; this may be useful to skip over invalid
        or expired tokens that exist, for example.

    **kwargs
        all keyword arguments are passed on to
        :meth:`scitokens.SciToken.deserialize`

    Returns
    -------
    token : `scitokens.SciToken`
        the first token that matches the requirements

    Raises
    ------
    ~igwn_auth_utils.IgwnAuthError
        if no valid token can be found

    See also
    --------
    scitokens.SciToken.deserialize
        for details of the deserialisation, and any valid keyword arguments
    """
    # iterate over all of the tokens we can find for this audience
    for token in _find_tokens(audience=audience, **kwargs):
        # parsing a token yielded an exception, handle it here:
        if isinstance(token, Exception):
            if skip_errors:
                continue
            raise token

        # if this token is valid, stop here and return it
        if is_valid_token(token, audience, scope, timeleft):
            return token

    # if we didn't find any valid tokens:
    raise IgwnAuthError(
        "could not find a valid SciToken, "
        "please verify the audience and scope, "
        "or generate a new token and try again",
    )


def _find_tokens(**deserialize_kwargs):
    """Yield all tokens that we can find

    This function will `yield` exceptions that are raised when
    attempting to parse a token that was actually found, so that
    they can be handled by the caller.
    """
    from scitokens.utils.errors import SciTokensException

    def _token_or_exception(func, *args, **kwargs):
        try:
            return func(*args, **kwargs)
        except SciTokensException as exc:
            return exc

    # read token directly from 'SCITOKEN{_FILE}' variable
    for envvar, loader in (
        ('SCITOKEN', deserialize_token),
        ('SCITOKEN_FILE', load_token_file),
    ):
        if envvar in os.environ:
            yield _token_or_exception(
                loader,
                os.environ[envvar],
                **deserialize_kwargs,
            )

    try:
        yield _token_or_exception(SciToken.discover, **deserialize_kwargs)
    except (
        OSError,  # no token
        AttributeError,  # windows doesn't have geteuid
    ) as exc:
        if isinstance(exc, AttributeError) and not (
            WINDOWS
            and "geteuid" in str(exc)
        ):
            raise
        for tokenfile in _find_condor_creds_token_paths():
            yield _token_or_exception(
                load_token_file,
                tokenfile,
                **deserialize_kwargs,
            )


def _find_condor_creds_token_paths():
    """Find all token files in the condor creds directory
    """
    try:
        _condor_creds_dir = Path(os.environ["_CONDOR_CREDS"])
    except KeyError:
        return
    try:
        for f in _condor_creds_dir.iterdir():
            if f.suffix == ".use":
                yield f
    except FileNotFoundError:   # creds dir doesn't exist
        return


# -- HTTP request helper ----

def token_authorization_header(token, scheme="Bearer"):
    """Format an in-memory token for use in an HTTP Authorization Header.

    Parameters
    ----------
    token : `scitokens.SciToken`
        the token to format

    scheme : `str` optional
        the Authorization scheme to use

    Returns
    -------
    header_str : `str`
        formatted content for an `Authorization` header

    Notes
    -----
    See `RFC-6750 <https://datatracker.ietf.org/doc/html/rfc6750>`__
    for details on the ``Bearer`` Authorization token standard.
    """
    return "{} {}".format(
        scheme,
        token._serialized_token or token.serialize().decode("utf-8"),
    )



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Tests for :mod:`igwn_auth_utils.x509`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import os
from datetime import (
    datetime,
    timedelta,
)
from pathlib import Path
from unittest import mock

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import (
    hashes,
    serialization,
)

import pytest

from .. import x509 as igwn_x509
from ..error import IgwnAuthError


# -- fixtures ---------------

@pytest.fixture
def x509cert(private_key, public_key):
    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, "test"),
    ])
    now = datetime.utcnow()
    return x509.CertificateBuilder(
        issuer_name=name,
        subject_name=name,
        public_key=public_key,
        serial_number=1000,
        not_valid_before=now,
        not_valid_after=now + timedelta(seconds=86400),
    ).sign(private_key, hashes.SHA256())


def _write_x509(cert, path):
    with open(path, "wb") as file:
        file.write(cert.public_bytes(
            serialization.Encoding.PEM,
        ))


@pytest.fixture
def x509cert_path(x509cert, tmp_path):
    cert_path = tmp_path / "x509.pem"
    _write_x509(x509cert, cert_path)
    return cert_path


# -- tests ------------------

@mock.patch.dict("os.environ")
@mock.patch(
    "os.getlogin" if os.name == "nt" else "os.getuid",
    mock.MagicMock(return_value=123),
)
def test_default_cert_path():
    if os.name == "nt":
        os.environ["SYSTEMROOT"] = r"C:\WINDOWS"
        expected = r"C:\WINDOWS\Temp\x509up_123"
    else:
        expected = r"/tmp/x509up_u123"
    assert igwn_x509._default_cert_path() == Path(expected)


def test_validate_certificate(x509cert):
    igwn_x509.validate_certificate(x509cert)


def test_validate_certificate_path(x509cert_path):
    igwn_x509.validate_certificate(x509cert_path)


def test_validate_certificate_expiry_error(x509cert):
    with pytest.raises(ValueError) as exc:
        igwn_x509.validate_certificate(x509cert, timeleft=int(1e10))
    assert str(exc.value) == (
        "X.509 certificate has less than 10000000000 seconds remaining"
    )


def test_is_valid_certificate(x509cert_path):
    assert igwn_x509.is_valid_certificate(x509cert_path)


def test_is_valid_certificate_false(tmp_path):
    assert not igwn_x509.is_valid_certificate(tmp_path / "does-not-exist")


@mock.patch.dict("os.environ")
def test_find_credentials_x509usercertkey(x509cert_path, public_pem_path):
    """Test that `find_credentials()` returns the X509_USER_{CERT,KEY} pair
    """
    os.environ.pop("X509_USER_PROXY", "test")  # make sure this doesn't win
    # configure the environment to return (cert, key)
    x509cert_filename = str(x509cert_path)
    x509key_filename = str(public_pem_path)
    os.environ["X509_USER_CERT"] = x509cert_filename
    os.environ["X509_USER_KEY"] = x509key_filename

    # check that find_credentials() returns the (cert, key) pair
    assert igwn_x509.find_credentials() == (
        x509cert_filename,
        x509key_filename,
    )


@mock.patch.dict("os.environ")
def test_find_credentials_x509userproxy(x509cert_path):
    """Test that `find_credentials()` returns the X509_USER_PROXY if set

    ... if X509_USER_{CERT,KEY} are not set
    """
    # remove CERT,KEY so that PROXY can win
    os.environ.pop("X509_USER_CERT", None)
    os.environ.pop("X509_USER_KEY", None)
    # set the PROXY variable
    x509cert_filename = str(x509cert_path)
    os.environ["X509_USER_PROXY"] = x509cert_filename
    # make sure it gets found
    assert igwn_x509.find_credentials() == x509cert_filename


@mock.patch.dict("os.environ", clear=True)
@mock.patch("igwn_auth_utils.x509._default_cert_path")
def test_find_credentials_default(_default_cert_path, x509cert_path):
    """Test that `find_credentials()` returns the default path

    ... if none of the X509_USER variable are set
    """
    _default_cert_path.return_value = x509cert_path
    assert igwn_x509.find_credentials() == str(x509cert_path)


@mock.patch.dict(
    "os.environ",
)
@mock.patch(
    "igwn_auth_utils.x509.is_valid_certificate",
    side_effect=(
        False,  # fail for _default_cert_path
        True,  # so that ~/.globus/usercert.pem passes
    ),
)
@mock.patch("os.access", return_value=True)
def test_find_credentials_globus(_, x509cert_path):
    """Test that .globus files are returned if all else fails
    """
    # clear X509 variables out of the environment
    for suffix in ("PROXY", "CERT", "KEY"):
        os.environ.pop(f"X509_USER_{suffix}", None)

    # check that .globus is found
    globusdir = Path.home() / ".globus"
    assert igwn_x509.find_credentials() == (
        str(globusdir / "usercert.pem"),
        str(globusdir / "userkey.pem"),
    )


@mock.patch.dict("os.environ")
@mock.patch("igwn_auth_utils.x509.is_valid_certificate", return_value=False)
def test_find_credentials_error(_):
    """Test that a failure in discovering X.509 creds raises the right error
    """
    # clear X509 variables out of the environment
    for suffix in ("PROXY", "CERT", "KEY"):
        os.environ.pop(f"X509_USER_{suffix}", None)

    # check that we can't find any credentials
    with pytest.raises(IgwnAuthError) as exc:
        igwn_x509.find_credentials()
    assert str(exc.value).startswith(
        "could not find an RFC-3820 compliant X.509 credential",
    )



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

import os
from datetime import datetime
from pathlib import Path

from cryptography.x509 import (
    Certificate,
    load_pem_x509_certificate,
)
from cryptography.hazmat.backends import default_backend

from .error import IgwnAuthError


def load_x509_certificate_file(file, backend=None):
    """Load a PEM-format X.509 certificate from a file, or file path

    Parameters
    ----------
    file : `str`, `pathlib.Path`, `file`
        file object or path to read from

    backend : `module`, optional
        the `cryptography` backend to use

    Returns
    -------
    cert : `cryptography.x509.Certificate`
        the X.509 certificate
    """
    if isinstance(file, (str, bytes, os.PathLike)):
        with open(file, "rb") as fileobj:
            return load_x509_certificate_file(fileobj)
    if backend is None:  # cryptography < 3.1 requires a non-None backend
        backend = default_backend()
    return load_pem_x509_certificate(file.read(), backend=backend)


def validate_certificate(cert, timeleft=600):
    """Validate an X.509 certificate by checking it's expiry time

    Parameters
    ----------
    cert : `cryptography.x509.Certificate`, `str`, `file`
        the certificate object or file (object or path) to validate

    timeleft : `float`, optional
        the minimum required time until expiry (seconds)

    Raises
    ------
    ValueError
        if the certificate has expired or is about to expire
    """
    # load a certificate from a file
    if not isinstance(cert, Certificate):
        cert = load_x509_certificate_file(cert)

    # then validate it
    if _timeleft(cert) < timeleft:
        raise ValueError(
            f"X.509 certificate has less than {timeleft} seconds remaining"
        )


def is_valid_certificate(cert, timeleft=600):
    """Returns True if ``cert`` contains a valid X.509 certificate

    Parameters
    ----------
    cert : `cryptography.x509.Certificate`, `str`, `file`
        the certificate object or file (object or path) to validate

    timeleft : `float`, optional
        the minimum required time until expiry (seconds)

    Returns
    -------
    isvalid : `bool`
        whether the certificate is valid
    """
    try:
        validate_certificate(cert, timeleft=timeleft)
    except (
        OSError,  # file doesn't exist or isn't readable
        ValueError,  # cannot load PEM certificate or expiry looming
    ):
        return False
    return True


def _timeleft(cert):
    """Returns the time remaining (in seconds) for a ``cert``
    """
    return (cert.not_valid_after - datetime.utcnow()).total_seconds()


def _default_cert_path(prefix="x509up_"):
    """Returns the temporary path for a user's X509 certificate

    Examples
    --------
    On Windows:

    >>> _default_cert_path()
    'C:\\Users\\user\\AppData\\Local\\Temp\\x509up_user'

    On Unix:

    >>> _default_cert_path()
    '/tmp/x509up_u1000'
    """
    if os.name == "nt":  # Windows
        tmpdir = Path(os.environ["SYSTEMROOT"]) / "Temp"
        user = os.getlogin()
    else:  # Unix
        tmpdir = "/tmp"
        user = "u{}".format(os.getuid())
    return Path(tmpdir) / "{}{}".format(prefix, user)


def find_credentials(timeleft=600):
    """Locate X509 certificate and (optionally) private key files.

    This function checks the following paths in order:

    - ``${X509_USER_CERT}`` and ``${X509_USER_KEY}``
    - ``${X509_USER_PROXY}``
    - ``/tmp/x509up_u${UID}``
    - ``~/.globus/usercert.pem`` and ``~/.globus/userkey.pem``

    Note
    ----
    If the ``X509_USER_{CERT,KEY,PROXY}`` variables are set, their paths
    **are not** validated in any way, but are trusted to point at valid,
    non-expired credentials.
    The default paths in `/tmp` and `~/.globus` **are** validated before
    being returned.

    Parameters
    ----------
    timeleft : `int`
        minimum required time left until expiry (in seconds)
        for a certificate to be considered 'valid'

    Returns
    -------
    cert : `str`
        the path of the certificate file that also contains the
        private key, **OR**

    cert, key : `str`
        the paths of the separate cert and private key files

    Raises
    ------
    ~igwn_auth_utils.IgwnAuthError
        if not certificate files can be found, or if the files found on
        disk cannot be validtted.

    Examples
    --------
    If no environment variables are set, but a short-lived certificate has
    been generated in the default location:

    >>> find_credentials()
    '/tmp/x509up_u1000'

    If a long-lived (grid) certificate has been downloaded:

    >>> find_credentials()
    ('/home/me/.globus/usercert.pem', '/home/me/.globus/userkey.pem')
    """
    # -- check the environment variables (without validation)

    try:
        return os.environ['X509_USER_CERT'], os.environ['X509_USER_KEY']
    except KeyError:
        try:
            return os.environ['X509_USER_PROXY']
        except KeyError:
            pass

    # -- look up some default paths (with validation)

    # 1: /tmp/x509up_u<uid> (cert = key)
    default = str(_default_cert_path())
    if is_valid_certificate(default, timeleft):
        return default

    # 2: ~/.globus/user{cert,key}.pem
    try:
        globusdir = Path.home() / ".globus"
    except RuntimeError:  # pragma: no cover
        # no 'home'
        pass
    else:
        cert = str(globusdir / "usercert.pem")
        key = str(globusdir / "userkey.pem")
        if (
            is_valid_certificate(cert, timeleft)  # validate the cert
            and os.access(key, os.R_OK)  # sanity check the key
        ):
            return cert, key

    raise IgwnAuthError(
        "could not find an RFC-3820 compliant X.509 credential, "
        "please generate one and try again.",
    )



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Tests for :mod:`igwn_auth_utils.requests`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Leo Singer <leo.singer@ligo.org>"

import os
import stat
from unittest import mock

import pytest

try:
    from .. import requests as igwn_requests
except ModuleNotFoundError as exc:  # pragma: no cover
    pytest.skip(str(exc), allow_module_level=True)
else:
    from requests import RequestException
from ..error import IgwnAuthError
from .test_scitokens import rtoken  # noqa: F401


def _empty(*args, **kwargs):
    return []


def _igwnerror(*args, **kwargs):
    raise IgwnAuthError("error")


def mock_no_scitoken():
    return mock.patch(
        "igwn_auth_utils.scitokens._find_tokens",
        _empty,
    )


def mock_no_x509():
    return mock.patch(
        "igwn_auth_utils.requests.find_x509_credentials",
        _igwnerror,
    )


@pytest.fixture
def netrc(tmp_path):
    netrc = tmp_path / "netrc"
    netrc.write_text(
        "machine example.org login albert.einstein password super-secret",
    )
    netrc.chmod(stat.S_IRWXU)
    return netrc


def has_auth(session):
    return bool(
        session.auth
        or session.cert
        or "Authorization" in session.headers
    )


class TestSession:
    Session = igwn_requests.Session

    # -- SessionErrorMixin

    def test_raise_for_status_hook(self, requests_mock):
        # define a request that returns 404 (not found)
        requests_mock.get(
            "https://test.org",
            status_code=404,
            reason="not found",
        )

        # with the kwarg a RequestException is raised
        with pytest.raises(RequestException) as exc:
            igwn_requests.get("https://test.org")
        assert str(exc.value) == (
            "404 Client Error: not found for url: https://test.org/"
        )

    # -- SessionAuthMixin

    def test_noauth_args(self):
        """Test that `Session(force_noauth=True, fail_if_noauth=True)` is invalid
        """
        with pytest.raises(ValueError):
            self.Session(force_noauth=True, fail_if_noauth=True)

    def test_fail_if_noauth(self):
        """Test that `Session(fail_if_noauth=True)` raises an error
        """
        with pytest.raises(IgwnAuthError):
            self.Session(
                token=False,
                cert=False,
                url=None,
                fail_if_noauth=True,
            )

    def test_force_noauth(self):
        """Test that `Session(force_noauth=True)` overrides auth kwargs
        """
        sess = self.Session(cert="cert.pem", force_noauth=True)
        assert not has_auth(sess)

    @mock_no_scitoken()
    @mock_no_x509()
    def test_defaults(self):
        """Test that the `Session()` defaults work in a noauth environment
        """
        sess = self.Session()
        assert not has_auth(sess)

    # -- tokens

    def test_token_explicit(self, rtoken):  # noqa: F811
        """Test that tokens are handled properly
        """
        sess = self.Session(token=rtoken)
        assert sess.headers["Authorization"] == (
            igwn_requests.scitoken_authorization_header(rtoken)
        )
        assert not sess.cert  # only one cred type is stored

    def test_token_serialized(self, rtoken):  # noqa: F811
        """Test that serialized tokens are handled properly
        """
        serialized = rtoken.serialize()
        sess = self.Session(token=serialized)
        assert sess.headers["Authorization"] == f"Bearer {serialized}"

    @mock.patch("igwn_auth_utils.requests.find_scitoken")
    def test_token_discovery(self, find_token, rtoken):  # noqa: F811
        find_token.return_value = rtoken
        sess = self.Session()
        assert sess.headers["Authorization"] == (
            igwn_requests.scitoken_authorization_header(rtoken)
        )

    @mock.patch(
        "igwn_auth_utils.requests.find_scitoken",
        side_effect=IgwnAuthError,
    )
    def test_token_required_failure(self, _):
        with pytest.raises(IgwnAuthError):
            self.Session(token=True)

    # -- X.509

    def test_cert_explicit(self):
        """Test that cert credentials are stored properly
        """
        sess = self.Session(token=False, cert="cert.pem")
        assert sess.cert == "cert.pem"
        assert has_auth(sess)

    @mock.patch(
        "igwn_auth_utils.requests.find_x509_credentials",
        return_value="test.pem",
    )
    def test_cert_discovery(self, _):
        """Test that automatic certificate discovery works
        """
        assert self.Session(token=False).cert == "test.pem"

    @mock.patch(
        "igwn_auth_utils.requests.find_x509_credentials",
        side_effect=IgwnAuthError,
    )
    def test_cert_required_failure(self, _):
        with pytest.raises(IgwnAuthError):
            self.Session(token=False, cert=True)

    # -- basic auth

    @mock.patch.dict(os.environ)
    @pytest.mark.parametrize(("url", "auth"), (
        ("https://example.org", ("albert.einstein", "super-secret")),
        ("https://bad.org", None),
    ))
    def test_basic_auth(self, netrc, url, auth):
        os.environ["NETRC"] = str(netrc)
        sess = self.Session(cert=False, token=False, url=url)
        assert sess.auth == auth


def test_get(requests_mock):
    """Test that `igwn_auth_utils.requests.get` can perform a simple request
    """
    requests_mock.get(
        "https://test.org",
        text="TEST",
    )
    assert igwn_requests.get("https://test.org").text == "TEST"


@mock.patch("igwn_auth_utils.requests.Session")
def test_get_session(mock_session):
    """Test that ``session`` for `igwn_auth_utils.requests.get` works
    """
    session = mock.MagicMock()
    assert igwn_requests.get("https://test.org", session=session)
    session.get.assert_called_once_with("https://test.org")
    mock_session.assert_not_called()



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Python Requests interface with IGWN authentication

This is heavily inspired by Leo Singer's excellent
:mod:`requests_gracedb` package.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Leo Singer <leo.singer@ligo.org>"

from urllib.parse import urlparse

from safe_netrc import netrc

import requests

from . import (
    IgwnAuthError,
    find_scitoken,
    find_x509_credentials,
    scitoken_authorization_header,
)

_auth_session_parameters = """
    Discovery/configuration of authorisation/authentication methods
    is attempted in the following order:

    1.  if ``force_noauth=True`` is given, no auth is configured;

    2.  if a bearer token is provided via the ``token`` keyword argument,
        then use that;

    3.  if an X.509 credential path is provided via the ``cert`` keyword
        argument, then use that;

    4.  if ``auth`` keyword is provided, then use that;

    5.  look for a bearer token by passing the ``token_audience``
        and ``token_scope`` keyword parameters to
        :func:`igwn_auth_utils.find_scitokens`;

    6.  look for an X.509 credential using
        :func:`igwn_auth_utils.find_x509_credential`

    7.  read the netrc file located at :file:`~/.netrc`, or at the path
        stored in the :envvar:`NETRC` environment variable, and look
        for a username and password matching the hostname given in the
        ``url`` keyword argument;

    8.  if none of the above yield a credential, and ``fail_if_noauth=True``
        was provided, raise a `ValueError`.

    Parameters
    ----------
    token : `scitokens.SciToken`, `str`, `bool`, optional
        Bearer token (scitoken) input, one of

        - a bearer token (`scitokens.SciToken`),
        - a serialised token (`str`, `bytes`),
        - `False`: disable using tokens completely
        - `True`: discover a valid token via
          :func:`igwn_auth_utils.find_scitoken` and
          error if something goes wrong
        - `None`: try and discover a valid token, but
          try something else if that fails

    token_audience, token_scope : `str`
        The ``audience`` and ``scope`` to pass to
        :func:`igwn_auth_utils.find_scitoken` when discovering
        available tokens.

    cert : `str`, `tuple`, `bool`, optional
        X.509 credential input, one of

        - path to a PEM-format certificate file,
        - a ``(cert, key)`` `tuple`,
        - `False`: disable using X.509 completely
        - `True`: discover a valid cert via
          :func:`igwn_auth_utils.find_x509_credentials` and
          error if something goes wrong
        - `None`: try and discover a valid cert, but
          try something else if that fails

    auth :  `tuple`, `object`, optional
        ``(username, password)`` `tuple` or other authentication/authorization
        object to attach to a `~requests.Request`

    url : `str`, optional
        the URL that will be queried within this session; this is only
        used to access credentials via :mod:`safe_netrc`.

    force_noauth : `bool`, optional
        Disable the use of any authorisation credentials (mainly for testing).

    fail_if_noauth : `bool`, optional
        Raise a `~igwn_auth_utils.IgwnAuthError` if no authorisation
        credentials are presented or discovered.

    Raises
    ------
    IgwnAuthError
        If ``cert=True`` or ``token=True`` is given and the relevant
        credential was not actually discovered, or
        if ``fail_if_noauth=True`` is given and no authorisation
        token/credentials of any valid type are presented or discovered.

    Note
    ----
    This class requires `Requests <https://requests.readthedocs.io/>`__.

    See also
    --------
    requests.Session
        for details of the standard options
    """.strip()


def _find_cred(func, *args, error=True, **kwargs):
    """Find a credential and maybe ignore an `~igwn_auth_utils.IgwnAuthError`

    This is an internal utility for the `SessionAuthMixin._init_auth`
    method which shouldn't necessary fail if it doesn't
    find a credential of any one type, but should just move on to the
    next option.
    """
    try:
        return func(*args, **kwargs)
    except IgwnAuthError:
        if error:
            raise
        return


def _hook_raise_for_status(response, *args, **kwargs):
    """Response hook to raise exception for any HTTP error (status >= 400)

    Reproduced (with permission) from :mod:`requests_gracedb.errors`,
    authored by Leo Singer.
    """
    return response.raise_for_status()


class SessionErrorMixin:
    """A mixin for :class:`requests.Session` to raise exceptions for HTTP
    errors.

    Reproduced (with permission) from :mod:`requests_gracedb.errors`,
    authored by Leo Singer.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hooks.setdefault("response", []).append(
            _hook_raise_for_status,
        )


class SessionAuthMixin:
    """Mixin for :class:`requests.Session` to add support for IGWN auth.

    By default this mixin will automatically attempt to discover/configure
    a bearer token (scitoken) or an X.509 credential, with options to
    require/disable either of those, or all authentication entirely.

    {parameters}
    """
    def __init__(
        self,
        token=None,
        token_audience=None,
        token_scope=None,
        cert=None,
        auth=None,
        url=None,
        force_noauth=False,
        fail_if_noauth=False,
        **kwargs,
    ):
        # initialise session
        super().__init__(**kwargs)

        # handle options
        if force_noauth and fail_if_noauth:
            raise ValueError(
                "cannot select both force_noauth and fail_if_noauth",
            )
        if force_noauth:
            return

        # find creds if we can
        _auth = self._init_auth(
            auth=auth,
            cert=cert,
            token=token,
            token_audience=token_audience,
            token_scope=token_scope,
            url=url,
        )

        # if no auth was found, and we need it, fail here
        if not _auth and fail_if_noauth:
            raise IgwnAuthError("no valid authorisation credentials found")

    def _init_auth(
        self,
        auth=None,
        cert=None,
        token=None,
        token_audience=None,
        token_scope=None,
        url=None,
    ):
        # -- user-provided objects
        # the user gave us something specific, so always use it
        dummy = (None, False, True)
        if token not in dummy:
            return self._set_token_header(token)

        for attr, obj in (
                ("cert", cert),
                ("auth", auth),
        ):
            if obj not in dummy:
                setattr(self, attr, obj)
                return getattr(self, attr)

        # -- the user didn't give us something specific,
        #    so we go around the houses trying to find something
        #    we can use

        # bearer token
        if token is not False:  # not disabled
            token = self._find_token(
                token_audience,
                token_scope,
                error=bool(token),
            )
        if token:
            return self._set_token_header(token)

        # cert auth
        if cert is not False:  # not disabled
            self.cert = self._find_x509_credentials(error=bool(cert))
        if self.cert:
            return self.cert

        # basic auth (netrc)
        if auth is not False:  # not disabled
            self.auth = self._find_username_password(url)

        return self.auth

    def _set_token_header(self, token):
        """Serialise a `scitokens.SciToken` and format and store as an
        Authorization header for this session.

        Parameters
        ----------
        token : `scitokens.SciToken`, `str`, `bytes`
            the token to serialize, or an already serialized representation
        """
        if isinstance(token, (str, bytes)):  # load a valid token
            header = f"Bearer {token}"
        else:
            header = scitoken_authorization_header(token)
        self.headers["Authorization"] = header
        return header

    @staticmethod
    def _find_x509_credentials(error=True):
        """Find an X.509 certificate for authorization
        """
        return _find_cred(find_x509_credentials, error=error)

    @staticmethod
    def _find_token(audience, scope, error=True):
        """Find a bearer token for authorization
        """
        return _find_cred(find_scitoken, audience, scope, error=error)

    @staticmethod
    def _find_username_password(url):
        """Use `safe_netrc.netrc` to find the username/password for basic auth
        """
        host = urlparse(url).hostname

        try:
            result = netrc().authenticators(host)
        except IOError:
            return

        try:
            username, _, password = result
        except TypeError:
            return
        return username, password


class Session(
    SessionAuthMixin,
    SessionErrorMixin,
    requests.Session,
):
    """`requests.Session` class with default IGWN authorization handling

    {parameters}

    Examples
    --------
    >>> from igwn_auth_utils.requests import Session
    >>> with Session(token=mytoken) as sess:
    ...     sess.get("https://my.science.stuff/data")
    """


# update the docstrings to include the same parameter info
for _obj in (Session, SessionAuthMixin):
    _obj.__doc__ = _obj.__doc__.format(parameters=_auth_session_parameters)


def get(url, *args, session=None, **kwargs):
    """Request data from a URL via an HTTP ``'GET'`` request

    Parameters
    ----------
    url : `str`,
        the URL to request

    session : `requests.Session`, optional
        the connection session to use, if not given one will be
        created on-the-fly

    *args, **kwargs
        all other keyword arguments are passed directly to
        `requests.Session.get`

    Returns
    -------
    resp : `requets.Response`
        the response object

    See also
    --------
    requests.Session.get
        for information on how the request is performed
    """
    # user's session
    if session:
        return session.get(url, *args, **kwargs)

    # new session
    sess_kwargs = {k: kwargs.pop(k, None) for k in (
        "cert",
        "token",
        "token_audience",
        "token_scope",
    )}
    with Session(url=url, **sess_kwargs) as session:
        return session.get(url, *args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Utilities for the `igwn_auth_utils` test suite
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key

import pytest


@pytest.fixture(scope="session")  # one per suite is fine
def private_key():
    return generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )


@pytest.fixture(scope="session")
def public_key(private_key):
    return private_key.public_key()


@pytest.fixture(scope="session")
def public_pem(public_key):
    return public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )


@pytest.fixture()
def public_pem_path(public_pem, tmp_path):
    pem_path = tmp_path / "key.pem"
    with open(pem_path, "wb") as file:
        file.write(public_pem)
    return pem_path



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from setuptools import setup
setup(use_scm_version=True)
